
import os
import re
import sys
import time
import urllib
import signal
import socket
import urllib2
import datetime
from xml.dom.minidom import parse
from xml.sax.saxutils import XMLGenerator

import cherrypy

import cherrypy._cpengine

from graphtool.base.xml_config import XmlConfig

class TfcRule:
  
  def __init__( self, mapper ):
    self.chain = None 
    self.protocol = None
    self.destination_match = None
    self.path_match = None 
    self.result = None
    self.mapper = mapper
    self.type = None
    
  def set( self, destination_match = None, path_match = None ):
    if destination_match != None:
      self.destination_match = re.compile( destination_match )
      self.destination_match_str = destination_match
    if path_match != None:
      self.path_match_str = path_match
      self.path_match = re.compile( path_match )
      
  def doesMatch( self, input, protocol, dest="" ):
    if protocol != self.protocol:
      return False 
    if self.destination_match and re.match( self.destination_match, dest ) == None:
      return False
    #if re.match( self.path_match, input ) == None:
    #  return False
    return True
    
  def map( self, input, protocol, dest="" ):
    orig_input = str(input)
    if not self.doesMatch(input, protocol, dest):
      return None
    if self.type == 'to_pfn':
      input = self.do_chain( input, protocol, dest )
      if input == None:
        raise ValueError("Could not map input %s using current TFC rules." % orig_input)
    my_match = re.match( self.path_match, input )
    if not my_match:
        #raise ValueError("Could not match input %s to rule %s." % (input, self.path_match_str))
        return None
    groups = my_match.groups()
    current_string = self.result
    for group, ind in zip(groups, range(len(groups)) ):
      pattern_str = '\$' + str(ind+1)
      sub_pattern = re.compile( pattern_str )
      current_string = re.sub( sub_pattern, group, current_string )
    if self.type == 'to_lfn':
      current_string = self.do_chain( current_string, protocol, dest )
    return current_string

  def do_chain( self, input, prot, dest ):
    #print "Doing chain", input, prot, self.chain
    if self.chain != None:
      if self.type == 'to_pfn':
        map = self.mapper.map_to_pfn
      else:
        map = self.mapper.map_to_lfn
      real_input = map( input, self.chain, dest )
    else:
      real_input = input
    return real_input

class TfcMapper(object):

  def __init__( self, dom=None ):
    self.to_pfn_rules = []
    self.to_lfn_rules = []
    if dom != None:
      self.start_parse( dom )

  def start_parse( self, dom ):
    sm = dom.getElementsByTagName('storage-mapping')
    if len(sm) > 1:
      raise ValueError("Only one storage mapping permitted per TFC.")
    if len(sm) == 0:
      raise ValueError("No storage mapping present!")
    self.parse_sm( sm[0] )

  def parse_sm( self, sm ):
    lfn_to_pfns = sm.getElementsByTagName('lfn-to-pfn')
    for lfn_to_pfn in lfn_to_pfns:
      self.parse_lfn_to_pfn( lfn_to_pfn )

    pfn_to_lfns = sm.getElementsByTagName('pfn-to-lfn')
    for pfn_to_lfn in pfn_to_lfns:
      self.parse_pfn_to_lfn( pfn_to_lfn )

  def parse_lfn_to_pfn( self, lfn_to_pfn ):
    rule = self.parse_rule( lfn_to_pfn )
    rule.type = "to_pfn"
    self.to_pfn_rules.append( rule )

  def parse_pfn_to_lfn( self, pfn_to_lfn ):
    rule = self.parse_rule( pfn_to_lfn )
    rule.type = "to_lfn"
    self.to_lfn_rules.append( rule )

  def parse_rule( self, rule ):
    my_rule = TfcRule( self )
    my_rule.protocol = rule.getAttribute('protocol')
    my_rule.set( destination_match = rule.getAttribute('destination-match') )
    my_rule.set( path_match = rule.getAttribute('path-match') )
    if len(rule.getAttribute('chain')) > 0:
      my_rule.chain = rule.getAttribute('chain')
    my_rule.result = rule.getAttribute('result')
    return my_rule

  def map_to_pfn( self, input, protocol, dest="" ):
    return self.map( input, protocol, dest, self.to_pfn_rules )

  def map_to_lfn( self, input, protocol, dest="" ):
    return self.map( input, protocol, dest, self.to_lfn_rules )

  def map( self, input, protocol, dest, rules ):
    result = None
    for rule in rules:
      result = rule.map( input, protocol, dest )
      if result != None:
        return result

class CheckException(Exception):
    pass

class CmsQuery(XmlConfig):

    _cpconfig = {'tools.caching.on': False}

    def query_dbsdiscovery(self, string, dbsInst='cms_dbs_prod_global', nostep=False):
        endpoint = self.metadata.get('dbsdiscovery', 'https://cmsweb.cern.ch/' \
            'dbs_discovery/aSearch')
        data = {'dbsInst': dbsInst, 'html': '0', 'xml': '1', 'caseSensitive': \
            '0', 'limit': '1', 'userInput': string}
        if nostep:
            data['pagerStep'] = 0
            data['limit'] = -1
        query = urllib.urlencode(data)
        fp = urllib2.urlopen(endpoint, query)
        return parse(fp)

    def getTfc(self, site=None):
        if site == None:
            site = self.getCmsName()
        url = self.metadata.get('tfc_mapper', 'http://t2.unl.edu/phedex/tfc/' \
            'xml')
        data = urllib.urlencode({'node': site})
        #fp = urllib2.urlopen(url, data)
        fp = urllib2.urlopen(url + "?" + data)
        mapper = TfcMapper()
        mapper.start_parse(parse(fp))
        return mapper

    def check(self, filename):
        self.check_namespace(filename)
        self.check_read(filename)
    check.exposed=True

    def check_namespace(self, filename):
        if not os.path.exists(filename):
            raise CheckException("Namespace entry does not exist.")

    def check_read(self, filename):
        timeout = float(self.metadata.get('timeout', '10'))
        args = ['dcap_sip', '-t', str(timeout), filename]
        pid = os.spawnvp(os.P_NOWAIT, 'dcap_sip', args)
        #print pid, args
        timer = -time.time()
        while timer + time.time() < timeout:
           try:
               status = os.waitpid(pid, os.P_NOWAIT)
           except:
               raise
           #print status
           if status[1]:
               #print status
               raise CheckException("Unable to read file via dcap.")
           if status[0]:
               return
           time.sleep(.05)
        #print pid
        try:
            os.kill(pid, signal.SIGKILL)
        except:
            pass
        try:
            os.wait()
        except:
            pass
        raise CheckException("DCap read timed out after %.2f seconds." % \
            timeout)

    def guessSite(self):
        try:
            domain = socket.getfqdn().split('.', 1)[1]
        except:
            domain = socket.getfqdn()
        dom = self.query_dbsdiscovery("find site where site=*.%s" % domain)
        #start = datetime.datetime(1970, 1, 1, 0, 0, 0)
        #curDate = start
        numProcDs = -1
        sitename = ''
        for row_dom in dom.getElementsByTagName('row'):
            name = str(row_dom.getElementsByTagName('sename')[0].firstChild.\
                data)
            #creationstr = str(row_dom.getElementsByTagName('creationdate')[0] \
            #    .firstChild.data)
            #creationtuple = time.strptime(creationstr, '%d %b %Y %H:%M:%S GMT')
            #creationdate = datetime.datetime(*creationtuple[:6])
            curProcDs = int(row_dom.getElementsByTagName('numberofprocds')[0] \
                .firstChild.data)
            #print name, creationdate
            #if creationdate > curDate:
            #    curDate = creationdate
            #    sitename = name
            if curProcDs > numProcDs:
                numProcDs = curProcDs
                sitename = name
        #if curDate == start:
        if numProcDs == -1:
            raise Exception("Unable to determine site name.")
        return sitename

    def getSite(self):
        try:
            return self.guessSite()
        except:
            raise
            return socket.getfqdn()
    getSite.exposed = True

    def getCmsName(self):
        return self.getCmsSite(self.getSite())
    getCmsName.exposed = True

    def getCmsSite(self, site):
        sitedb = self.metadata.get('sitedb_xml', 'https://cmsweb.cern.ch/' \
            'sitedb/sitedb/xml/index')
        url = os.path.join(sitedb, 'SEtoCMSName')
        data = urllib.urlencode({'name': site})
        dom = parse(urllib2.urlopen(url, data))
        for name_dom in dom.getElementsByTagName('name'):
            return str(name_dom.firstChild.data)
        raise Exception("Unable to determine site name!")

    def checkCmsDataset(self, *args, **kw):
        dsName = os.path.join(*(['/'] + list(args)))
        if kw.get('site', None) == None:
            site = self.getSite()
        else:
            site = kw['site']
        string = 'find file where dataset=%s and site=%s' % (dsName, site)
        return self.checkCmsFromDBS(string, site)
    checkCmsDataset.exposed = True

    def checkCmsBlock(self, *args, **kw):
        bName = os.path.join(*(['/'] + list(args)))
        if kw.get('block', None) == None:
            pass
        else:
            bName = kw['block']
        if '#' not in bName:
            raise Exception("No hash (#) located in block name; did you" \
                " forget to replace # with %23 in the URL?")
        if kw.get('site', None) == None:
            site = self.getSite()
        else:
            site = kw['site']
        string = 'find file where block=%s and site=%s' % (bName, site)
        return self.checkCmsFromDBS(string, site)
    checkCmsBlock.exposed = True

    def checkCmsFromDBS(self, string, site):
        cms_site = self.getCmsSite(site)
        dom = self.query_dbsdiscovery(string, nostep=True)
        tfc = self.getTfc(cms_site)
        lfns = []
        for row in dom.getElementsByTagName('row'):
            lfn = str(row.getElementsByTagName('logicalfilename')[0].\
                firstChild.data)
            lfns.append(lfn)
        return self.checkCmsFiles(lfns, tfc)

    def checkCmsFile(self, *args, **kw):
        fileName = os.path.join(*(['/'] + list(args)))
        if kw.get('site', None) == None:
            site = self.getSite()
        cms_site = self.getCmsSite(site)
        tfc = self.getTfc(cms_site)
        return self.checkCmsFiles([fileName], tfc)
    checkCmsFile.exposed = True

    def checkCmsFiles(self, lfns, tfc):
        def genFunc():
            for lfn in lfns:
                file = tfc.map_to_pfn(lfn, protocol='direct')
                try:
                    self.check(file)
                    msg = "OK"
                except CheckException, ce:
                    msg = "FAIL: " + str(ce)
                #print "in generator", lfn, msg
                #raise Exception("Blah")
                yield "%s %s\n" % (lfn, msg)
        cherrypy.response.headers['Content-Type'] = 'text/plain'
        cherrypy.response.stream = True
        return genFunc()

