
import os
import sys
import optparse
import signal

# Author: Chad J. Schroeder
# Copyright: Copyright (C) 2005 Chad J. Schroeder
UMASK = 0
WORKDIR = "/"
MAXFD = 1024

if (hasattr(os, "devnull")):
   REDIRECT_TO = os.devnull
else:
   REDIRECT_TO = "/dev/null"

def createDaemon(cmd, args=[], cmdName=None):
   """Detach a process from the controlling terminal and run it in the
   background as a daemon.
   """

   try:
      pid = os.fork()
   except OSError, e:
      raise Exception, "%s [%d]" % (e.strerror, e.errno)

   if (pid == 0):	# The first child.
      os.setsid()
      try:
         pid = os.fork()	# Fork a second child.
      except OSError, e:
         raise Exception, "%s [%d]" % (e.strerror, e.errno)

      if (pid == 0):	# The second child.
         os.chdir(WORKDIR)
         os.umask(UMASK)
         print "Launched %s" % cmdName
         for i in range(3):
             os.close(i)
         os.open(REDIRECT_TO, os.O_RDWR)      # standard input (0)
         os.dup2(0, 1)                        # standard output (1)
         os.dup2(0, 2)                        # standard error (2)
         try:
             fp = open('/var/run/%s.pid' % cmdName, 'w')
             fp.write(str(os.getpid()))
             fp.close()
         except:
             pass
         cmd_args = cmd.split()[1:]
         cmd =cmd.split()[0]
         os.execv(cmd, [cmdName] + cmd_args + args)
      else:
         os._exit(0)	# Exit parent (the first child) of the second child.
   else:
      os._exit(0)	# Exit parent of the first child.

   import resource		# Resource usage information.
   maxfd = resource.getrlimit(resource.RLIMIT_NOFILE)[1]
   if (maxfd == resource.RLIM_INFINITY):
      maxfd = MAXFD
  
   for fd in range(0, maxfd):
      try:
         os.close(fd)
      except OSError:	# ERROR, fd wasn't open to begin with (ignored)
         pass

   os.open(REDIRECT_TO, os.O_RDWR)	# standard input (0)

   os.dup2(0, 1)			# standard output (1)
   os.dup2(0, 2)			# standard error (2)

   return(0)

def parseArgs():
    usage = "usage: %prog [options] action"
    parser = optparse.OptionParser(usage=usage)
    parser.add_option("-u", "--user", dest="user", help="Start up program as " \
        "USER", metavar="USER", default=None)
    parser.add_option('-c', '--cmd', dest='cmd', help="Start CMD.", \
        metavar="CMD", default=None)
    parser.add_option('-n', '--cmdName', dest='name', help="Command's name.", \
        default=None)
    parser.add_option("-a", "--args", dest="args", help="Arguments to start " \
        "the command.", default="")
    options, args = parser.parse_args()
    if len(args) != 1:
        parser.print_help()
        print "Need to specify exactly one action (start/stop/status)."
        sys.exit(1)
    return options, args, parser

def pid_is_running(pid):
    pid = int(pid)
    try:
        os.getpgid(pid)
        return True
    except:
        return False

def stop(name):
    try:
        old_pid = int(open('/var/run/%s.pid' % name, 'r').read())
    except:
        print "Unable to determine running process ID for %s." % name
        sys.exit(1)
    if pid_is_running(old_pid):
        os.kill(int(old_pid), signal.SIGTERM)
        print "%s (pid: %s) has been sent the terminate signal." % (name, old_pid)
    else:
        print "%s (pid: %s) does not appear to be running." % (name, old_pid)
        sys.exit(1)

def status(name):
    try:
        old_pid = int(open('/var/run/%s.pid' % name, 'r').read())
    except:
        print "Unable to determine running process ID for %s." % name
        sys.exit(1)
    if pid_is_running(old_pid):
        print "%s (pid: %s) is running." % (name, old_pid)
    else:
        print "%s is not running." % name
    sys.exit(0)

def start(cmd, args=None, user=None, cmdName=None):
    try:
        old_pid = int(open('/var/run/%s.pid' % name, 'r').read())
    except:
        old_pid=None
    if old_pid and pid_is_running(old_pid):
        print "%s (pid: %i) is already running." % (name, old_pid)
        sys.exit(2)
    if not cmdName:
        cmdName = cmd.split()[0]
        cmdName = cmd.split('/')[-1]
    if user:
        cmd = '/usr/bin/sudo -u %s %s' % (user, cmd)
    createDaemon(cmd, args.split(), cmdName)

def main():
    options, args, parser = parseArgs()
    if not options.name:
        parser.print_help()
        print "Must specify a command name."
    if args[0] == 'start':
        if not options.cmd:
            parser.print_help()
            print "Need to specify the command to start."
            sys.exit(1)
        start(options.cmd, options.args, options.user, options.name)
    if args[0] == 'stop':
        stop(options.name)
    if args[0] == 'status':
        status(options.name)

if __name__ == '__main__':
    main()

